import numpy as np
from typing import Union


def create_mesh(*args: [list[int]]):
    r"""Creates a mesh centered around zero with the dimensions provided.

    :param args:    Integer dimensions.
    :return:        List of mesh grids with the provided dimensions.

    Example
    _____________

    To create a list of numpy arrays of a fleshed out meshgrid with dimensions (11, 8) with the x dimension ranging
    from -5 to 5 and y from -3.5 to 3.5, both in steps on 1.

    >>> [x, y] = create_mesh(11, 8)
    """

    if any(not isinstance(v, int) for v in args) or min(args) < 1:
        raise ValueError("Only positive integer values can be converted to a mesh.")
    vectors = map(lambda x: np.arange(x) - (x - 1) / 2, args)
    return np.meshgrid(*vectors, indexing='ij')


def cart2pol(x: Union[float, np.ndarray], y: Union[float, np.ndarray]) -> np.ndarray:
    """Converts cartesian coordinates to polar coordinates.

    :param x:   x coordinates.
    :param y:   y coordinates.
    :return:    Array of polar coordinates [radius, theta].
    """
    r = np.sqrt(x ** 2 + y ** 2)
    theta = np.arctan2(y, x)
    return np.array([r, theta])


def vortex(nx: int, ny: int) -> np.ndarray:
    """Creates a vortex with dimensions (nx, ny) at the center of the array.

    :param nx:  x size of the vortex.
    :param ny:  y size of the vortex.
    :return:    Vortex state with dimensions (3, x, y), where the first index runs over the components.
    """
    min_dim = min([nx, ny])
    xx, yy = create_mesh(nx, ny)
    rad, theta = cart2pol(xx, yy)

    mx = -np.cos(theta)
    my = np.sin(theta)
    mz = np.zeros_like(theta)

    mx = np.where(rad > (min_dim - 1) / 2, 0, mx)
    my = np.where(rad > (min_dim - 1) / 2, 0, my)

    return np.array([np.zeros_like(mx), mx, my, mz])


def skyrmion(nx: int, ny: int, number: int = 1, helicity: float = 0, polarity: int = 1,
             neel: bool = False) -> np.ndarray:
    """Creates a skyrmion state with dimensions (nx, ny) at the center of the array.

    :param nx:          x size.
    :param ny:          y size.
    :param number:      Skyrmion topological number.
    :param helicity:    Helicity offset.
    :param polarity:    mz magnetization at the center of the skyrmion.
    :param neel:        True for Neel type skyrmion, False (default) for Bloch type.
    :return:            Skyrmion state with dimensions (3, x, y), where the first index runs over components.
    """

    xx, yy = create_mesh(nx, ny)
    azimuth = np.arctan2(xx, -yy)
    rad = np.sqrt(xx ** 2 + yy ** 2)
    # normalize theta to equal pi at the edges
    theta = 2 * rad / (np.min([nx, ny]) - 1) * np.pi

    mx = np.sin(theta) * np.cos(number * (azimuth + helicity))
    my = np.sin(theta) * np.sin(number * (azimuth + helicity))
    mz = polarity * np.cos(theta)

    if neel:
        mx, my = my, -mx
        mz = mz * neel * number

    mx = np.where(theta > np.pi, 0, mx)
    my = np.where(theta > np.pi, 0, my)
    mz = np.where(theta > np.pi, -polarity, mz)

    return np.array([np.zeros_like(mx), mx, my, mz])


def tessellate(unit_cell: np.ndarray, times: Union[int, list[int]], pattern: str = 'sq') -> np.ndarray:
    """Create a lattice from the given unit cell.

    :param unit_cell:       Structure to be tessellated, shaped (..., nx, ny)
    :param times:           The amount of times to tessellate structure in both dimensions
    :param pattern:         'hex' for hexagonal or 'sq' for square tiling
    :return:                Crustal structure shaped (..., nx*times, ny*times)
    """

    unit_shape = unit_cell.shape[1:3]
    nx, ny = unit_shape
    if type(times) is int:
        times = np.array([times, times])
    elif type(times) is list and len(times) == 2:
        times = np.array(times)
    else:
        print('Times must be an integer or a list of two integers.')
        return unit_cell

    if pattern[0:2] == ('sq' or 're'):
        return np.tile(unit_cell, times)
    elif pattern[0:3] == 'hex':
        if nx != ny:
            print('Hexagonal tiling only works with square unit cells.')
        final_shape = unit_shape * times
        times += 2
        filled = np.tile(unit_cell, times)
        select = [i + nx for i in range(final_shape[1] - nx) if i % (2 * nx) in range(nx)]
        partial_structure = filled[:, :, select]
        temp = np.roll(partial_structure, nx // 2, axis=1)
        filled[:, :, select] = temp
        return filled[:, nx:-nx, ny:-ny]
    else:
        print('Tesselation pattern not recognised.')
        return unit_cell


def circle_mask(structure: np.ndarray) -> np.ndarray:
    """Masks the structure with a circle of radius equal to the smallest dimension of the structure.

    :param structure:   Structure to be masked.
    :return:            Masked structure.
    """
    xx, yy = create_mesh(*structure.shape[-2:])
    boundary_values = [v[0, 0] for v in structure]
    rad = np.min([xx.max(), yy.max()])
    circle = (xx ** 2 + yy ** 2 < rad ** 2)
    for j in range(structure.shape[0]):
        structure[j] = structure[j] * circle + (1 - circle) * boundary_values[j]
    return structure


def pad_circle(structure: np.ndarray, times: int = 2, repeat_mode: str = 'edge') -> np.ndarray:
    """Applies a circular mask and then pads the structure with the given repeat mode.

    :param structure:       Structure to be padded.
    :param times:           The amount of padding as a multiple of the structure size.
    :param repeat_mode:     The mode to use for padding.
    :return:                Padded structure.
    """
    structure = circle_mask(structure)
    times = tuple([2 * (int(times * size),) for size in structure.shape[-2:]])
    if structure.ndim == 3:
        times = ((0, 0), *times)
        return np.pad(structure, times, mode=repeat_mode)
    elif structure.ndim == 2:
        return np.pad(structure, times, mode=repeat_mode)
    else:
        raise ValueError("Structure must be 2 or 3 dimensional.")


def gaussian_2d(coords: list[np.ndarray], sigma: list[float, ...]) -> np.ndarray:
    """Created a 2D gaussian centered around zero.

    :param coords:      The coordinate system list (e.g. np.meshgrid(...)).
    :param sigma:       A list of sigma values for the Gaussian.
    :return:            The 2D Gaussian centered around zero."""

    zz = [_gaussian(ii, s) for ii, s in zip(coords, sigma)]
    return np.multiply.reduce(zz)


def _gaussian(x: np.ndarray, sigma: float):
    """One-dimensional gaussian function."""
    if sigma <= 0:
        return np.ones_like(x)
    return np.exp(-(x ** 2 / (2 * sigma ** 2)))



import logging
import magpack.vectorop
import numpy as np
import copy
from typing import Union
import magneticScattering.dft as dft
import magneticScattering.structures as structures
from typing import cast


def _typecast_roi(roi):
    """Casts a roi array into tuple[float,float,float,float].

    Parameters
    ----------
    roi : array_like

    Returns
    -------
    4-tuple of float
        Type-casted region of interest / extent
    """
    return cast(tuple[float, float, float, float], roi)


def pauli():
    """Returns an array of the identity and three Pauli matrices.

    Returns
    -------
    np.ndarray
        Numpy array containing the identity and three Pauli matrices with shape (4, 2, 2).
    """
    p_matrix = np.zeros([4, 2, 2], dtype=np.complex128)
    p_matrix[0] = np.eye(2)
    p_matrix[1] = np.array([[1, 0], [0, -1]])
    p_matrix[2] = np.array([[0, 1], [1, 0]])
    p_matrix[3] = np.array([[0, -1j], [1j, 0]])
    return p_matrix


def en2wave(energy):
    """Converts energy in eV to wavelength in m.

    Parameters
    ----------
    energy : float
        Energy in electronvolts.

    Returns
    -------
    float
        Corresponding wavelength in m.
    """
    return 1.24e-6 / energy


def vec2np(vec, size=2, dtype=np.float64):
    """Converts a scalar or array value to an appropriately-sized numpy array.

    Parameters
    ----------
    vec : np.ndarray | list | tuple
        Scalar value or list of scalar values.
    size : int (optional)
        Desired size of array.
    dtype : np.dtype (optional)
        Desired dtype of array.

    Returns
    -------
    np.ndarray
        Numpy array with provided size and datatype.
    """
    if isinstance(vec, (tuple, list)):
        if len(vec) == 1:
            return np.array(vec * size, dtype=dtype)
        elif len(vec) == size:
            return np.array(vec, dtype=dtype)
        else:
            logging.error('Too many or not enough elements in list, expected {} got {}'.format(size, len(vec)))
            raise ValueError

    elif isinstance(vec, (int, float, complex)):
        return np.array([vec] * size, dtype=dtype)

    elif isinstance(vec, np.ndarray):
        assert vec.shape == (size,), 'Wrong shape of array, expected {} got {}'.format((size,), vec.shape)
        return vec


class Beam:
    """Describes the properties of the beam."""

    def __init__(self, wavelength, fwhm, pol):
        """
        Parameters
        ----------
        wavelength : float
            The wavelength of the beam in meters.
        fwhm : np.ndarray | 2-tuple | 2-list
            The full width at half maximum of the beam in the x and y directions.
        pol : 4-list of float
            The polarization of the beam in Stokes parameters.
        """
        self._wavelength = wavelength
        self._fwhm = vec2np(fwhm)
        self._beam_sigma = self._fwhm / np.sqrt(8 * np.log(2))
        self._pol = vec2np(pol, 4)
        self._density_matrix, self._degree_of_pol = self.calc_density_matrix()
        self._change_tracker = [lambda x: None]

    def calc_density_matrix(self):
        r"""Converts the Stokes polarization vector to a density matrix

        Returns
        -------
        np.ndarray
            The density matrix of the beam.
        np.ndarray
            The degree of polarization.

        Notes
        -----
        The expression for converting the Strokes polarization "vector" to a density matrix is:

        .. math::

            \rho = \frac{1}{2}(\vec{P} \cdot \vec{\sigma})


        where :math:`\vec{P}` is the Stokes polarization vector and :math:`\sigma` is the vector containing the Pauli spin
        matrices and the identity matrix :math:`(\sigma_0 = \mathbf{I})`.
        """
        pol_intensity = np.sum(self._pol[-3:] ** 2)
        pol_norm = self._pol[0] ** 2
        if pol_intensity > pol_norm:
            self._pol[-3:] = self._pol[-3:] / np.sqrt(pol_intensity)
            logging.warning("The intensity of the polarization vector is larger than the norm.")

        if pol_norm > 1:
            self._pol = self._pol / np.sqrt(pol_norm)
            logging.warning("The norm of the polarization vector is larger than 1.")

        density_matrix = np.einsum('i,ijk->jk', self._pol, pauli()) / 2
        degree_of_pol = np.einsum('aij,ji->a', pauli(), density_matrix)[-3:].real ** 2 * 100

        return density_matrix, degree_of_pol

    @property
    def tracker(self):
        """Function to notify parent of any changes to the class. """
        return self._change_tracker

    @tracker.setter
    def tracker(self, parent_tracker_function):
        self._change_tracker.append(parent_tracker_function)

    @property
    def wavelength(self):
        """The wavelength of the beam. """
        return self._wavelength

    @wavelength.setter
    def wavelength(self, wavelength: float):
        self._wavelength = wavelength
        [f('wavelength') for f in self._change_tracker]

    @property
    def fwhm(self):
        """The full width at half maximum of the beam. """
        return self._fwhm

    @fwhm.setter
    def fwhm(self, fwhm: Union[np.ndarray, list[float], float]):
        self._fwhm = vec2np(fwhm)
        self._beam_sigma = self._fwhm / np.sqrt(8 * np.log(2))
        [f('fwhm') for f in self._change_tracker]

    @property
    def sigma(self):
        """The spread of the beam. """
        return self._beam_sigma

    @property
    def pol(self):
        """The polarisation of the beam. """
        return self._pol

    @pol.setter
    def pol(self, pol: Union[np.ndarray, list[float]]):
        self._pol = vec2np(pol, 4)
        self._density_matrix, self._degree_of_pol = self.calc_density_matrix()
        [f('pol') for f in self._change_tracker]

    @property
    def density_matrix(self):
        """Polarization density matrix."""
        return self._density_matrix

    @property
    def degree_of_pol(self):
        """The degree of polarisation. """
        return self._degree_of_pol


class Sample:
    """Describes the sample properties."""

    def __init__(self, sample_length, scattering_factors, structure):
        """
        Parameters
        ----------
        sample_length : array_like
            The dimensions of the sample in meters.
        scattering_factors : 3-list of complex
            The complex scattering factors corresponding to charge, XMCD and XMLD.
        structure : np.ndarray
            The magnetic (and electronic) configuration of the sample.
        """
        self._sample_length = vec2np(sample_length)
        self._scattering_factors = vec2np(scattering_factors, 3, dtype=np.complex128)

        if structure.ndim != 3:
            raise ValueError("Structure has wrong shape, must be (3 or 4, x, y).")
        if structure.shape[0] == 3:
            # add a charge component in place of the magnetization
            charge = magpack.vectorop.magnitude(structure) > 0
            structure = np.concatenate([charge[np.newaxis, ...], structure], axis=0)
        elif structure.shape[0] != 4:
            raise ValueError("Structure has wrong shape, must be (3 or 4, x, y).")

        self._structure = copy.deepcopy(structure)
        self._pix_size = self.calc_pix_size()
        self._change_tracker = [lambda x: None]

    @property
    def tracker(self):
        """Function to notify parent of any changes to the class. """
        return self._change_tracker

    @tracker.setter
    def tracker(self, parent_tracker_function):
        self._change_tracker.append(parent_tracker_function)

    def calc_pix_size(self):
        """Calculates the size of a single pixel in the structure.

        Returns
        -------
        np.ndarray
            Pixel size along the x and y dimensions.
        """
        return np.divide(self._sample_length, self.shape)

    @property
    def sample_length(self):
        """The lateral size of the entire sample or array configuration. """
        return self._sample_length

    @sample_length.setter
    def sample_length(self, v: Union[np.ndarray, list[float], float]):
        self._sample_length = vec2np(v, 2)
        self._pix_size = self.calc_pix_size()
        [f('sample_length') for f in self._change_tracker]

    @property
    def scattering_factors(self):
        """The three complex scattering factors of the material."""
        return self._scattering_factors

    @scattering_factors.setter
    def scattering_factors(self, v: Union[list[complex], np.ndarray]):
        self._scattering_factors = vec2np(v, 3, dtype=np.complex128)
        [f('scattering_factors') for f in self._change_tracker]

    @property
    def shape(self):
        """The shape of the sample. """
        return self._structure.shape[-2:]

    @property
    def pix_size(self):
        """The pixel size of the sample; the size of a single pixel."""
        return self._pix_size

    @pix_size.setter
    def pix_size(self, v: Union[np.ndarray, list[float], float]):
        self._pix_size = vec2np(v, 2)
        self._sample_length = np.multiply(self._pix_size, self.shape)
        [f('pix_size') for f in self._change_tracker]

    @property
    def structure(self):
        """The charge and three magnetization components of the sample. """
        return self._structure

    @structure.setter
    def structure(self, v: np.ndarray):
        if v.ndim != 3 or v.shape[0] != 4:
            logging.error("Structure has wrong shape, must be (4, x, y).")
            raise ValueError
        self._structure = copy.deepcopy(v)
        self._pix_size = self.calc_pix_size()
        [f('structure') for f in self._change_tracker]

    def get_extent(self):
        """The extent of the sample for plotting.

        Returns
        -------
        np.ndarray
            Array with shape (4,) describing the extent of the sample for plotting."""
        extent_x = self.sample_length[0] * np.array([-0.5, 0.5])
        extent_y = self.sample_length[1] * np.array([-0.5, 0.5])
        return np.hstack([extent_x, extent_y])

    def get_coordinates(self):
        """Meshgrid spanning the sample.

        Returns
        -------
        list of np.ndarray
            list containing the coordinates obtained from meshgrid."""
        nx, ny = self.shape
        extent = self.get_extent()
        xx, yy = np.linspace(extent[0], extent[1], nx), np.linspace(extent[2], extent[3], ny)
        return list(np.meshgrid(xx, yy, indexing='ij'))


class Geometry:
    """Defines the geometry of the experiment."""

    def __init__(self, angle, detector_distance):
        """
        Parameters
        ----------
        angle : float
            The angle of incidence (and scattering) of the beam in degrees.
        detector_distance
            The distance between the sample and the detector in meters.
        """
        self._angle = angle
        self._detector_distance = detector_distance
        self._change_tracker = [lambda x: None]

    @property
    def tracker(self):
        """Function to notify parent of any changes to the class."""
        return self._change_tracker

    @tracker.setter
    def tracker(self, parent_tracker_function):
        self._change_tracker = parent_tracker_function

    @property
    def angle(self):
        """The angle of incidence of the beam [degrees]."""
        return self._angle

    @angle.setter
    def angle(self, v: float):
        self._angle = v
        [f('angle') for f in self._change_tracker]

    @property
    def detector_distance(self):
        """The distance between the sample and the detector [m]."""
        return self._detector_distance

    @detector_distance.setter
    def detector_distance(self, v: float):
        self._detector_distance = v
        [f('detector_distance') for f in self._change_tracker]


class Scatter:
    """Initializes the Scatter class and calculates the scattering pattern."""
    def __init__(self, beam, sample, geometry):
        """
        Attributes
        ----------
        beam : :class:`Beam`
            The properties of the beam.
        sample : :class:`Sample`
            The sample being measured.
        geometry : :class:`Geometry`
            The geometry of the experiment.
        """

        self.beam = beam
        self.beam.tracker = self.change_tracker
        self.sample = sample
        self.sample.tracker = self.change_tracker
        self.geometry = geometry
        self.geometry.tracker = self.change_tracker
        self._extent = self._get_extent()
        self._roi = self._extent
        self._roi_shape = self.sample.shape
        self.run()

    @property
    def roi_shape(self):
        """Shape of the region of interest."""
        return self._roi_shape

    @property
    def extent(self):
        """The real space occupied by the scattering pattern."""
        return self._extent

    @property
    def angular_extent(self):
        """The solid angle that the scattering pattern occupies."""
        return np.rad2deg(np.array(self._extent) / self.geometry.detector_distance)

    @property
    def roi(self):
        """Region of interest for calculating the scattering pattern."""
        return self._roi

    @roi.setter
    def roi(self, v: list[float]):
        v = vec2np(v, 4)
        self._roi = v
        self.run()

    @property
    def roi_angular_extent(self):
        """Solid angle that the selected region of interest occupies."""
        # angular extent in degrees around the center
        return np.rad2deg(np.array(self._roi) / self.geometry.detector_distance)

    def change_tracker(self, name):
        """Listens for any changes in any of the classes so that the scattering can be re-calculated."""
        logging.info("Detected change in subclass attribute: {}.".format(name))
        if name in ['wavelength', 'sample_length', 'pix_size', 'detector_distance']:
            self._roi = self._get_extent()
        elif name in ['pol', 'fwhm', 'scattering_factors', 'angle', 'structure']:
            self._roi = self._get_extent()
            self.run()
        else:
            logging.error("Unknown attribute change detected: {}.".format(name))

    def _get_extent(self):
        angular_space = 1 / (2 * self.sample.pix_size) * self.beam.wavelength
        real_space = angular_space * self.geometry.detector_distance
        roi = np.array([-real_space[0], real_space[0], -real_space[1], real_space[1]])
        return _typecast_roi(roi)

    def run(self):
        """Calculates the scattering pattern."""
        # convolution of the beam with the structure
        conv_structure = self.sample.structure * structures.gaussian_2d(self.sample.get_coordinates(),
                                                                        self.beam.sigma)
        structure_sf = self._calc_scattering_factor(conv_structure)
        fft_structure_sf = self._calc_ft(structure_sf)
        self._calc_intensity(fft_structure_sf)

    def _calc_scattering_factor(self, m):
        f_0, f_1, f_2 = self.sample.scattering_factors
        theta = np.deg2rad(self.geometry.angle)

        s = np.sin(theta)
        c = np.cos(theta)
        s2 = np.sin(2 * theta)
        c2 = np.cos(2 * theta)

        nx, ny = self.sample.shape
        f = np.zeros((2, 2, nx, ny), dtype=np.complex128)
        f[0, 0, :, :] = f_0 * m[0] + f_2 * m[1] ** 2
        f[0, 1, :, :] = -1j * f_1 * (m[3] * c - m[2] * s) + f_2 * m[1] * (m[2] * c + m[3] * s)
        f[1, 0, :, :] = -1j * f_1 * (m[2] * s - m[3] * c) + f_2 * m[1] * (m[2] * c - m[3] * s)
        f[1, 1, :, :] = f_0 * m[0] * c2 - 1j * f_1 * m[1] * s2 + f_2 * ((m[2] * c) ** 2 - (m[3] * s) ** 2)
        return f

    def _calc_ft(self, structure_sf):
        # initial scattering is done using fast Fourier transform.
        if np.allclose(self.extent, self.roi):
            fft_structure_sf = np.fft.fftshift(np.fft.fftn(structure_sf, axes=(-2, -1)), axes=(-2, -1))
        # if a roi is selected, perform a discrete Fourier transform.
        else:
            roi_shape = self._roi_shape
            offset, sampling = dft.calc_params_from_roi(self.extent, self.roi, roi_shape)
            fft_structure_sf = dft.dftn_axes(structure_sf, ft_pix_vec=(0, 0) + roi_shape, offset_vec=(0, 0) + offset,
                                             upsample_vec=(0, 0) + sampling, axes=(-2, -1))
            fft_structure_sf = np.fft.fftshift(fft_structure_sf, axes=(-2, -1))
        return fft_structure_sf

    def _calc_intensity(self, fft_sf):
        mu_prime = np.einsum('ijab,jk,lkab->ilab', fft_sf, self.beam.density_matrix, fft_sf.conjugate())
        self.intensity = np.abs(np.einsum('iiab->ab', mu_prime))
        self.pol_out = np.abs(np.einsum('ijk,kjab->iab', pauli(), mu_prime))



import numpy as np
from magneticScattering import scatter, plot
import matplotlib.pyplot as plt
from magneticScattering.holography import holography_reference
from importlib import resources
import magneticScattering.data
import gzip

pol_dict = {'LH': [1, 1, 0, 0], 'LV': [1, -1, 0, 0],
            'CL': [1, 0, 0, 1], 'CR': [1, 0, 0, -1]}


def dichroism():
    """Simulated the circular magnetic dichroism from a labyrinthine pattern.

    See :ref:`example` for a more detailed description."""
    energy = 706  # energy of the beam in eV
    fwhm = 20e-6  # full width at half maximum of the beam in meters
    angle = 0  # angle of incidence
    detector_distance = 50e-2  # sample-detector distance
    sample_length = 10e-6  # sample size
    scattering_factors = [1j + 1, 1j + 1, 1 + 1j]  # scattering factors
    reference_hole_size = 3 # size of reference hole in pixels

    # load magnetic configuration
    inp_file = resources.files(magneticScattering.data) / 'labyrinthine.npy.gz'

    # Open it using gzip
    with inp_file.open("rb") as raw_f:
        with gzip.GzipFile(fileobj=raw_f, mode="rb") as gz_f:
            mag_config = np.load(gz_f)

    # create the sample
    sample = scatter.Sample(sample_length, scattering_factors, mag_config)
    # characterise the two beams
    beam_cp = scatter.Beam(scatter.en2wave(energy), [fwhm, fwhm], pol_dict['CL'])
    beam_cl = scatter.Beam(scatter.en2wave(energy), [fwhm, fwhm], pol_dict['CR'])
    # describe the geometry
    geometry = scatter.Geometry(angle, detector_distance)
    s_cp = scatter.Scatter(beam_cp, sample, geometry)
    s_cl = scatter.Scatter(beam_cl, sample, geometry)

    # see the magnetic configuration of the sample
    plot.structure(sample)
    plt.show()

    # interactive plot for looking at features more closely
    returned_roi = plot.intensity_interactive(s_cp, log=True)
    s_cl.roi = returned_roi
    s_cp.roi = returned_roi
    plot.difference(s_cl, s_cp, log=True)
    plt.show()

    # add a reference hole and perform holography
    holography_reference(sample, reference_hole_size, 'xy')
    plot.structure(sample)

    # invert holography to recover structure
    plot.holography(s_cp, s_cl, log=True, recons_only=False)  # entire pattern
    plot.holography(s_cp, s_cl)  # only sample reconstruction
    plt.show()


if __name__ == '__main__':
    dichroism()



import numpy as np
import logging


def dft(data, ft_pix = None, offset = None, upsample = 1):
    """One dimensional discrete Fourier transform (DFT).

    Parameters
    ----------
    data : np.ndarray
        Input data to transform.
    ft_pix : int (optional)
        Number of Fourier transform points.
    offset : float (optional)
        Frequency offset to adjust the frequency grid.
    upsample : float (optional)
        Factor by which to upsample the frequency grid.

    Returns
    -------
    np.ndarray
        Array of the computed Fourier transform.
    """
    n_pix = data.shape[0]
    ft_pix = n_pix if ft_pix is None else ft_pix
    offset = 0 if offset is None else offset
    kernel = np.einsum('k,n->kn',
                       np.fft.fftfreq(ft_pix, upsample) - offset / ft_pix, np.arange(n_pix))
    return np.sum(data * np.exp(-2j * np.pi * kernel), axis=1)


def dftn_axes(data, ft_pix_vec=None, offset_vec=None, upsample_vec=None, axes=None):
    """Performs n-dimensional discrete Fourier transform along specified axes.

    Parameters
    ----------
    data : np.ndarray
        Input data to transform.
    ft_pix_vec : list of int (optional)
        Number of Fourier transform points, one element per axis.
    offset_vec : list of int (optional)
        Frequency offset to adjust the frequency grid, one element per axis.
    upsample_vec : list of float (optional)
        Factor by which to upsample the frequency grid, one element per axis.
    axes : list of int (optional)
        Axes along which to perform the Fourier transform.

    Returns
    -------
        Array of the computed Fourier transform.
    """
    n_pix_vec, ndim = data.shape, data.ndim
    if axes is None:
        axes = tuple(range(data.ndim))
    axes = np.array(axes)
    axes = np.where(axes < 0, ndim + axes, axes)

    ft_pix_vec, offset_vec, upsample_vec = _dft_properties(ft_pix_vec, n_pix_vec, ndim, offset_vec, upsample_vec)

    for dim, n_pix, ft_pix, offset, upsample in zip(range(data.ndim), data.shape, ft_pix_vec, offset_vec,
                                                    upsample_vec):
        # Here, we take the fft of the first dimension, send it to the back and iterate:
        # >>> abc..xyz -> bc..xyzA -> c...xyzAB -> ... -> zABC...XY -> ABC...XYZ
        # capital letters denote fourier domain
        if dim in axes:
            kernel = np.einsum('k,n->kn',
                               np.fft.fftfreq(ft_pix, upsample) - offset / ft_pix, np.arange(n_pix))
            data = np.einsum('kn,n...->...k', np.exp(-2j * np.pi * kernel), data)
        else:
            data = np.rollaxis(data, data.ndim - 1)
    return data


def _dft_properties(ft_pix_vec, n_pix_vec, ndim, offset_vec, upsample_vec):
    """Validates the dft properties for the desired dimension and convert scalars to lists when necessary.

    Parameters
    ----------
    ft_pix_vec : list of int (optional)
        Vector of Fourier transform points.
    n_pix_vec : list of int (optional)
        Shape of in initial image.
    ndim : int
        Number of spatial dimensions.
    offset_vec : list of float (optional)
        Vector of frequency offsets.
    upsample_vec : list of float (optional)
        Vector of upsampling factors.

    Returns
    -------
    ft_pix_vec : list of int (optional)
        Vector of Fourier transform points.
    offset_vec : list of float (optional)
        Vector of frequency offsets.
    upsample_vec : list of float (optional)
        Vector of upsampling factors.
    """

    ft_pix_vec = n_pix_vec if ft_pix_vec is None else ft_pix_vec
    offset_vec = 0 if offset_vec is None else offset_vec
    upsample_vec = 1 if upsample_vec is None else upsample_vec
    ft_pix_vec = (ft_pix_vec,) * ndim if isinstance(ft_pix_vec, (int, float)) else ft_pix_vec
    offset_vec = (offset_vec,) * ndim if isinstance(offset_vec, (int, float)) else offset_vec
    upsample_vec = (upsample_vec,) * ndim if isinstance(upsample_vec, (int, float)) else upsample_vec
    return ft_pix_vec, offset_vec, upsample_vec


def calc_params_from_roi(current, roi, datashape):
    """Given the current extent of the image and the desired roi to upsample, the appropriate offset and upsample values
    are provided.

    Parameters
    ----------
    current : 4-tuple of float
        Current extent of the image.
    roi : 4-tuple of float
        Desired roi.
    datashape : tuple of int
        Desired shape of the data.

    Returns
    -------
    offset, upsample : tuple of float
        Offset and upsample values that will yield the desired roi when the discrete Fourier transform is calculated.
    """
    shape = np.array(datashape)
    current = np.array(current)
    roi = np.array(roi)

    range_ext = current[1] - current[0], current[3] - current[2]
    range_roi = roi[1] - roi[0], roi[3] - roi[2]
    pix_size = range_ext / shape

    logging.debug("range_ext: %s, range_roi: %s", range_ext, range_roi)

    center_ext = np.array([(current[0] + current[1]), (current[2] + current[3])])
    center_roi = np.array([(roi[0] + roi[1]), (roi[2] + roi[3])])
    logging.debug("center_ext: %s, center_roi: %s", center_ext, center_roi)

    upsample = range_ext[0] / range_roi[0], range_ext[1] / range_roi[1]
    logging.debug("upsample: %s", upsample)
    logging.debug("pix_size: %s", pix_size)
    logging.debug("pix_size*shape (should be extent no?): %s", pix_size * shape)

    offset = tuple((center_ext - center_roi) / 2 / pix_size)
    logging.debug("offset: %s", offset)

    return offset, upsample



from magneticScattering.holography import invert_holography
from magneticScattering.scatter import *
from matplotlib import colors
from matplotlib.widgets import RectangleSelector
from matplotlib import pyplot as plt
from matplotlib.axes import Axes

units = {'k': 1e3, '': 1, 'm': 1e-3, 'µ': 1e-6, 'n': 1e-9, 'p': 1e-12, 'f': 1e-15}


def _lin_thresh_pow(data):
    """Calculates a linear threshold value for logarithmic plotting."""
    return 10 ** np.ceil(np.log10(np.cbrt(np.abs(data).max())))


def _choose_scale(roi):
    """Choose appropriate units for plotting based on the range of values.

    Parameters
    ----------
    roi : tuple of float
        Region of interest coordinates.

    Returns
    -------
    prefix : str
        Units that will be used for plotting.
    scaled_roi : 4-tuple of float
        Region of interest scaled to account for the new units.
    """

    # Compute the log10 scale difference in x and y directions
    dx, dy = np.log10([roi[1] - roi[0], roi[3] - roi[2]])
    log_scales = {np.log10(value): prefix for prefix, value in units.items()}

    # Find the closest order of magnitude (rounded to a multiple of 3)
    x_closest, y_closest = 3 * round(dx / 3), 3 * round(dy / 3)

    overall_closest = min(x_closest, y_closest)
    prefix = log_scales.get(overall_closest, '')

    # Scale the ROI using the chosen unit
    scaled_roi = np.array(roi) / units[prefix]

    return prefix, cast(tuple[float, float, float, float], scaled_roi)


def structure(struct: Sample, **kwargs):
    """Plot the components of magnetisation structure.

    Parameters
    ----------
    struct : Sample
        The sample to plot the charge and magnetisation components.

    Other Parameters
    ----------------
    kwargs : dict
        Keyword arguments for matplotlib imshow.
    """
    extent = struct.get_extent()
    prefix, scaled_roi = _choose_scale(extent)

    fig, ax = plt.subplots(1, 4)
    fig.suptitle("Structure")
    title_index = ["Charge", "$m_x$", "$m_y$", "$m_z$"]
    fig.supxlabel("$x$ / " + prefix + "m")
    fig.supylabel("$y$ / " + prefix + "m")

    for i, (ax_i, colormap) in enumerate(zip(ax.flatten(), ['gray', 'PiYG', 'PuOr', 'RdBu'])):
        cbar_val = np.max(np.abs([struct.structure[i].min(), struct.structure[i].max()])) if i in [1, 2, 3] else None
        vmin = -cbar_val if cbar_val is not None else 0
        color = ax_i.imshow(struct.structure[i].T, origin='lower', extent=scaled_roi, cmap=colormap,
                            vmin=vmin, vmax=cbar_val, **kwargs)
        plt.colorbar(color, ax=ax_i, location='bottom')
        ax_i.set_aspect('equal')
        ax_i.set_title(title_index[i])


def pol(scatter: Scatter, log=True, **kwargs) -> None:
    """Plot the polarization states of the scattered light.

    Parameters
    ----------
    scatter : Scatter
        Class containing the intensity to view the different polarization states.
    log : bool
        Set to True for logarithmic scale.

    Other Parameters
    ----------------
    kwargs : dict
        Keyword arguments for matplotlib imshow.
    """
    roi = scatter.roi
    prefix, scaled_roi = _choose_scale(roi)
    vmin, vmax = scatter.pol_out[-3:].min(), scatter.pol_out[-3:].max()
    norm = colors.SymLogNorm(vmin, vmin=vmin, vmax=vmax) if log else None

    fig, ax = plt.subplots(1, 3)
    fig.suptitle("Relative Polarization States")
    title_index = ["Horizontal", "Diagonal", "Circular"]
    fig.supxlabel("Detector Position $x$ / " + prefix + "m")
    fig.supylabel("Detector Position $y$ / " + prefix + "m")

    im = None
    for i, ax_i in enumerate(ax.flatten()):
        im = ax_i.imshow(scatter.pol_out[i + 1].T, origin='lower', extent=scaled_roi, norm=norm, **kwargs)
        ax_i.set_aspect('equal')
        ax_i.set_title(title_index[i])
    cbar_ax = fig.add_axes((0.15, 0.15, 0.7, 0.05))
    fig.colorbar(im, cax=cbar_ax, orientation='horizontal')


def difference(scatter_a: Scatter, scatter_b: Scatter, log: bool = False, cmap: str = 'Spectral', **kwargs) -> None:
    """Plot the difference between two scattering patterns.

    Parameters
    ----------
    scatter_a, scatter_b : Scatter
        The classes containing the two patterns for which the difference will be plotted.
    log : bool (optional)
        Boolean choice to plot in log scale.
    cmap : str (optional)
        Color map to use.

    Other Parameters
    ----------------
    kwargs : dict
        Keyword arguments for matplotlib imshow.
    """
    if np.any(scatter_a.roi != scatter_b.roi):
        raise ValueError("Diffraction geometries have different parameters.")
    extent = scatter_a.roi
    diff = scatter_a.intensity - scatter_b.intensity
    if np.all(diff == 0):
        raise ValueError("No dichroism.")

    prefix, scaled_roi = _choose_scale(extent)
    norm = colors.SymLogNorm(_lin_thresh_pow(diff)) if log else None

    fig, ax = plt.subplots(1, 1)
    colorscale = ax.imshow(diff.T, origin='lower', extent=scaled_roi, norm=norm, cmap=cmap, **kwargs)
    fig.colorbar(colorscale)
    ax.set_title("Intensity Difference")
    ax.set_xlabel("Detector Position $x_0$ / " + prefix + "m")
    ax.set_ylabel("Detector Position $y_0$ / " + prefix + "m")


def intensity_interactive(scatter: Scatter, log: bool = False, **kwargs) -> tuple[float, float, float, float]:
    """Interactive plot of the intensity difference that can be used for selecting regions to view in higher resolution.

    Parameters
    ----------
    scatter : Scatter
        The scatter class to plot the intensity for.
    log : bool (optional)
        Boolean choice to plot in log scale.

    Returns
    -------
    4-tuple of floats
        The interactively-selected region of interest coordinates.

    Other Parameters
    ----------------
    kwargs : dict
        Keyword arguments for matplotlib imshow.
    """
    selected_regions = []

    def select_callback(click, release):
        x_start, y_start = click.xdata, click.ydata
        x_end, y_end = release.xdata, release.ydata
        selected_regions.append([x_start, x_end, y_start, y_end])

    roi = scatter.roi
    selected_regions.append(roi)
    norm = colors.SymLogNorm(1) if log else None
    prefix, scaled_roi = _choose_scale(roi)
    logging.info(scaled_roi)
    intensity_array = copy.deepcopy(scatter.intensity)

    fig, ax = plt.subplots(1, 1)
    selector = RectangleSelector(
        ax, select_callback,
        useblit=True,
        minspanx=0, minspany=0,
        spancoords='data',
        interactive=True,
        props=dict(facecolor='None', edgecolor='red'))
    fig.suptitle("Select region to view in higher resolution:")
    _intensity_image(ax, fig, intensity_array, norm, scaled_roi, prefix, prefix, **kwargs)
    fig.canvas.mpl_connect('key_press_event', selector)
    plt.show()

    x1, x2, y1, y2 = selected_regions[-1]

    return x1 * units[prefix], x2 * units[prefix], y1 * units[prefix], y2 * units[prefix]


def _intensity_image(ax: Axes, fig, intensity_array, norm, scaled_roi, x_prefix, y_prefix, **kwargs):
    """Adds appropriate axes labels to the intensity image."""
    colorscale = ax.imshow(intensity_array.T, origin='lower', extent=scaled_roi, norm=norm, **kwargs)
    fig.colorbar(colorscale)
    ax.set_title("Intensity")
    ax.set_xlabel("Detector Position $x_0$ / " + x_prefix + "m")
    ax.set_ylabel("Detector Position $y_0$ / " + y_prefix + "m")
    ax.axis('scaled')


def intensity(scatter, log = False, **kwargs):
    """Plot the intensity of the scattered light.

    Parameters
    ----------
    scatter : Scatter
        The scatter class whose intensity will be plotted.
    log : bool (optional)
        Boolean choice to plot in log scale.

    Other Parameters
    ----------------
    kwargs : dict
        Keyword arguments passed to matplotlib imshow function.
    """
    roi = scatter.roi
    norm = colors.SymLogNorm(1) if log else None
    prefix, scaled_roi = _choose_scale(roi)
    intensity_array = copy.deepcopy(scatter.intensity)
    fig, ax = plt.subplots(1, 1)
    _intensity_image(ax, fig, intensity_array, norm, scaled_roi, prefix, prefix, **kwargs)


def holography(scatter_a, scatter_b=None, log = False, recons_only=True,cmap=None, **kwargs):
    """Plot the result of inverting the holography pattern.

    Parameters
    ----------
    scatter_a : Scatter
        The scatter class whose intensity will be plotted.
    scatter_b : Scatter (optional)
        Optional second scatter class for plotting the difference.
    log : bool (optional)
        Boolean choice to plot in log scale.
    recons_only : bool (optional)
        When True, focus on the reconstruction on the sample and not the entire inverse scattering.
    cmap : str (optional)
        Colormap for the image

    Other Parameters
    ----------------
    kwargs : dict
    """
    inverse, roi = invert_holography(scatter_a, scatter_b)
    if recons_only:
        sx, sy = inverse.shape
        if sx > 3 * sy:
            inverse = inverse[0:sx // 4, :]  # reference hole along x
            roi[0], roi[1] = roi[0] / 4, roi[1] / 4

        elif sy > 3 * sx:
            inverse = inverse[:, 0:sy // 4]  # reference hole along y
            roi[2], roi[3] = roi[2] / 4, roi[3] / 4

        else:
            inverse = inverse[0:sx // 4, sy // 2 - sy // 8: sy // 2 + sy // 8]  # reference hole along x and y
            roi = [i / 4 for i in roi]


    if cmap is None:
        if scatter_b is None:
            cmap = 'viridis'
        elif cmap is None:
            cmap = 'Spectral'

    prefix, scaled_roi = _choose_scale(roi)
    norm = colors.SymLogNorm(_lin_thresh_pow(inverse)) if log else None

    fig, ax = plt.subplots(1, 1)
    ax.set_title("Reconstructed FTH")
    ax.set_xlabel("Relative Real Size $x_0$ / " + prefix + "m")
    ax.set_ylabel("Relative Real Size $y_0$ / " + prefix + "m")
    colorscale = ax.imshow(inverse.T, origin='lower', extent=scaled_roi, norm=norm, cmap=cmap, **kwargs)
    fig.colorbar(colorscale)

def scale_data(data, scale):
    """Scale the values in the array so that the minimum is scale[0] and maximum is scale[1].

    Parameters
    ----------
    data : np.ndarray
        Array to be scaled.
    scale : 2-list of float
        Scaling range as a list: [lower, upper].

    Returns
    -------
    np.ndarray
        Scaled array.
    """
    data = data - np.min(data)  # 0 to max+min
    data = data / np.max(data)  # 0 to 1
    data = data * (scale[1] - scale[0]) + scale[0]  # scale[0] to scale[1]
    return data



import magpack.image_utils
from magneticScattering.scatter import Sample, Scatter
import numpy as np


def holography_reference(sample, hole_size=1, axis='y'):
    """Creates a holography reference hole for the given sample.

    The new sample has a fourfold increased size along the specified axis to allow for recovery.

    Parameters
    ----------
    sample : np.ndarray | Sample
        The sample to add holography reference holes.
    hole_size : int (optional)
        The size of reference hole in voxels.
    axis : {'x', 'y', 'xy', 'yx'} (optional)
        Axis along which to add the reference.

    Returns
    -------
    np.ndarray (optional)
        The holography sample or None if directly applied to a Sample class.
    """
    if isinstance(sample, Sample):
        initial_structure = np.array(sample.structure)
    else:
        initial_structure = sample

    if hole_size <= 0:
        hole_size = 1

    if axis in ['xy', 'yx']:
        if isinstance(sample, Sample):
            holography_reference(sample, hole_size, 'x')
            holography_reference(sample, hole_size, 'y')
            return None
        else:
            step2 = holography_reference(holography_reference(initial_structure, hole_size, axis='x'), hole_size,
                                         axis='y')
            return step2
    elif axis == 'x':
        initial_structure = initial_structure.transpose(0, 2, 1)
    else:
        axis = 'y'

    components, init_x, init_y = initial_structure.shape

    # Create a zero array that is 4x4 times the size of the tile
    grid = np.zeros((components, init_x, init_y * 4))

    # Place the tile in the center
    center_x, center_y = init_x // 2, 2 * init_y  # True center
    grid[:, center_x - init_x // 2:center_x + init_x // 2,
    center_y - init_y // 2:center_y + init_y // 2] = initial_structure

    ref_x_center, ref_y_center = int(center_x), int(init_y * 10.5 / 3)
    # structure placed in the lower right corner
    grid[0, ref_x_center - hole_size // 2:ref_x_center + hole_size // 2 + 1,
    ref_y_center - hole_size // 2:ref_y_center + hole_size // 2 + 1] = 1

    if axis == 'x':
        grid = grid.transpose(0, 2, 1)

    if isinstance(sample, Sample):
        sample.sample_length = sample.sample_length * [1, 4] if axis == 'y' else sample.sample_length * [4, 1]
        sample.structure = grid

    else:
        return grid


def invert_holography(scatter_a, scatter_b=None):
    """Inverts the scattering pattern assuming Fourier transform holography was performed.

    Parameters
    ----------
    scatter_a : Scatter
        The scatter pattern to invert.
    scatter_b : Scatter (optional)
        The scatter pattern to invert for taking the difference.

    Returns
    -------
    inv : np.ndarray
        The inverted scatter pattern.
    roi : tuple of float
        The corresponding real space dimensions.
    """

    intensity1, len1 = scatter_a.intensity, scatter_a.sample.sample_length / 2
    if scatter_b is not None:
        intensity2, len2 = scatter_b.intensity, scatter_b.sample.sample_length / 2
    else:
        intensity2, len2 = None, None

    roi1 = [-len1[0], len1[0], -len1[1], len1[1]]

    if intensity2 is not None and intensity1.shape != intensity2.shape:
        raise ValueError("Scattering intensities must have the same shape.")
    if len1 is not None and len2 is not None and np.any(len1 != len2):
        raise ValueError("Scattering regions must be similar.")

    inv1 = np.abs(magpack.image_utils.fft(intensity1))
    inv2 = np.abs(magpack.image_utils.fft(intensity2)) if intensity2 is not None else None

    if inv2 is not None:
        return inv1 - inv2, roi1
    else:
        return inv1, roi1
